/*
 * (C) Copyright 2016 Nuxeo SA (http://nuxeo.com/) and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
package org.nuxeo.ecm.platform.importer.mqueues.pattern.producer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.nuxeo.ecm.core.api.Blob;
import org.nuxeo.ecm.core.api.Blobs;
import org.nuxeo.ecm.platform.importer.mqueues.pattern.message.DocumentMessage;
import org.nuxeo.ecm.platform.importer.random.HunspellDictionaryHolder;
import org.nuxeo.ecm.platform.importer.random.RandomTextGenerator;

import java.io.Serializable;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Random;
import java.util.concurrent.ThreadLocalRandom;

/**
 * @since 9.1
 */
public class RandomDocumentMessageProducer extends AbstractProducer<DocumentMessage> {
    private static final Log log = LogFactory.getLog(RandomDocumentMessageProducer.class);
    private final long nbDocuments;
    private final RandomBlobInfoProvider blobInfoProvider;
    private boolean countFolderAsDocument = true;
    private int maxFoldersPerFolder = 50;
    private int maxDocumentsPerFolder = 500;
    private int blobSizeKB = 0;
    private boolean blobOnlyText = false;

    private int documentCount = 0;
    private int folderCount = 0;
    private final Random rand;
    private static RandomTextGenerator gen;

    static protected final String[] DC_NATURE = {"article", "acknowledgement", "assessment", "application", "order",
            "contract", "quotation", "fax", "worksheet", "letter", "memo", "note", "notification", "procedure",
            "report", "internshipReport", "pressReview"};

    static protected final String[] DC_SUBJECTS = {"art/architecture", "art/comics", "art/cinema", "art/culture", "art/danse",
            "art/music", "sciences/astronomy", "sciences/biology", "sciences/chemistry", "sciences/math",
            "sciences/physic", "society/ecology", "daily life/gastronomy", "daily life/gardening", "daily life/sport",
            "technology/it"};

    static protected final String[] DC_RIGHTS = {"OpenContentL", "CC-BY-NC", "CC-BY-ND", "FreeArt", "ODbi", "GNUGPL",
            "FreeBSD", "CC0"};

    static protected final String[] DC_LANGUAGE = {"IT", "DE", "FR", "US", "EN"};

    static protected final String[] DC_SOURCE = {"internal", "external", "unknown"};

    static protected final String[] DC_COVERAGE = {"europe/France", "europe/Germany", "europe/Italy", "europe/Spain",
            "oceania/Tonga", "africa/Mali", "asia/Japan", "north-america/United_States_of_America"};
    private int foldersInCurrentFolderLimit;
    private int documentInCurrentFolderLimit;

    private enum DocType {Root, Folder, Document}

    private DocType currentType = DocType.Root;
    private int parentIndex = 0;
    private List<String> parents = new ArrayList<>();
    private List<String> children = new ArrayList<>();
    private int documentInCurrentFolderCount = 0;

    public RandomDocumentMessageProducer(int producerId, long nbDocuments, String lang, Path blobInfoDirectory) {
        super(producerId);
        this.nbDocuments = nbDocuments;
        rand = ThreadLocalRandom.current();

        synchronized (RandomDocumentMessageProducer.class) {
            if (gen == null) {
                gen = new RandomTextGenerator(new HunspellDictionaryHolder(lang));
                gen.prefilCache();
            }
        }
        if (blobInfoDirectory != null) {
            this.blobInfoProvider = new RandomBlobInfoProvider(blobInfoDirectory, producerId);
        } else {
            this.blobInfoProvider = null;
        }
        log.info("RandomDocumentMessageProducer created, nbDocuments: " + nbDocuments);
    }

    public RandomDocumentMessageProducer setMaxFoldersPerFolder(int max) {
        maxFoldersPerFolder = max;
        return this;
    }

    public RandomDocumentMessageProducer setMaxDocumentsPerFolder(int max) {
        maxDocumentsPerFolder = max;
        return this;
    }

    public RandomDocumentMessageProducer countFolderAsDocument(boolean value) {
        countFolderAsDocument = value;
        return this;
    }

    public RandomDocumentMessageProducer withBlob(int sizeKB, boolean onlyText) {
        this.blobSizeKB = sizeKB;
        this.blobOnlyText = onlyText;
        return this;
    }


    @Override
    public int getPartition(DocumentMessage message, int partitions) {
        return getProducerId() % partitions;
    }

    @Override
    public boolean hasNext() {
        if (countFolderAsDocument) {
            return (documentCount + folderCount) < nbDocuments;
        }
        return documentCount <= nbDocuments;
    }

    @Override
    public DocumentMessage next() {
        DocumentMessage ret;
        switch (currentType) {
            case Root:
                ret = createRoot();
                parents.add(ret.getId());
                currentType = DocType.Folder;
                foldersInCurrentFolderLimit = rand.nextInt(maxFoldersPerFolder) + 1;
                break;
            case Folder:
                ret = createFolder(parents.get(parentIndex));
                children.add(ret.getId());
                if (children.size() >= foldersInCurrentFolderLimit) {
                    currentType = DocType.Document;
                    documentInCurrentFolderCount = 0;
                    documentInCurrentFolderLimit = rand.nextInt(maxDocumentsPerFolder);
                }
                break;
            default:
            case Document:
                ret = createDocument(parents.get(parentIndex), children);
                documentInCurrentFolderCount += 1;
                if (documentInCurrentFolderCount > documentInCurrentFolderLimit) {
                    parentIndex += 1;
                    if (parentIndex >= parents.size()) {
                        parents.clear();
                        parents = children;
                        children = new ArrayList<>();
                        parentIndex = 0;
                    }
                    currentType = DocType.Folder;
                    foldersInCurrentFolderLimit = rand.nextInt(maxFoldersPerFolder) + 1;
                }
                break;
        }
        // log.debug(ret.getType() + ": " + ret.getId());
        return ret;
    }

    private DocumentMessage createRoot() {
        folderCount++;
        return getRandomNodeWithPrefix(String.format("%02d-", getProducerId()), "Folder", "");
    }


    private DocumentMessage createFolder(String parentPath) {
        DocumentMessage node = getRandomNode("Folder", parentPath, false);
        folderCount++;
        return node;
    }

    private DocumentMessage createDocument(String parentPath, List<String> exclude) {
        DocumentMessage node = getRandomNode("File", parentPath, true);
        String ret = node.getId();
        while (exclude.contains(ret)) {
            log.debug("duplicate found");
            node = getRandomNode("File", parentPath, true);
            ret = node.getId();
        }
        documentCount++;
        return node;
    }

    private DocumentMessage getRandomNode(String type, String parentPath, boolean withBlob) {
        String title = getTitle();
        String name = getName(title);
        HashMap<String, Serializable> props = getRandomProperties(title);
        DocumentMessage.Builder builder = DocumentMessage.builder(type, parentPath, name).setProperties(props);
        if (withBlob) {
            if (blobInfoProvider != null) {
                builder.setBlobInfo(blobInfoProvider.getBlobInfo(builder));
            } else {
                builder.setBlob(getRandomBlob());
            }
        }
        return builder.build();
    }

    private DocumentMessage getRandomNodeWithPrefix(String prefix, String type, String parentPath) {
        String title = getTitle();
        String name = prefix + getName(title);
        HashMap<String, Serializable> props = getRandomProperties(title);
        DocumentMessage.Builder builder = DocumentMessage.builder(type, parentPath, name).setProperties(props);
        if (blobInfoProvider != null) {
            builder.setBlobInfo(blobInfoProvider.getBlobInfo(builder));
        } else {
            builder.setBlob(getRandomBlob());
        }
        return builder.build();
    }

    private Blob getRandomBlob() {
        if (blobSizeKB == 0) {
            return null;
        }
        String content = gen.getRandomText(blobSizeKB);
        return Blobs.createBlob(content, getBlobMimeType(), null, getName(getTitle()) + ".txt");
    }

    private String getBlobMimeType() {
        if (blobOnlyText) {
            return "text/plain";
        } else {
            return "text/partial";
        }
    }

    private String getName(String title) {
        return title.replaceAll("\\W+", "-").toLowerCase();
    }

    private String getTitle() {
        return capitalize(gen.getRandomTitle(rand.nextInt(3) + 1).trim());
        //  return "f" + folderCount;
    }

    private String capitalize(final String line) {
        return Character.toUpperCase(line.charAt(0)) + line.substring(1);
    }

    protected HashMap<String, Serializable> getRandomProperties(String title) {
        HashMap<String, Serializable> ret = new HashMap<>();
        ret.put("dc:title", title);
        if (rand.nextInt(10) == 1) {
            String description = gen.getRandomTitle(rand.nextInt(5) + 1);
            ret.put("dc:description", capitalize(description));
        }
        ret.put("dc:nature", getGaussian(DC_NATURE));
        ret.put("dc:subjects", (Serializable) Collections.singletonList(getGaussian(DC_SUBJECTS)));
        ret.put("dc:rights", getGaussian(DC_RIGHTS));
        ret.put("dc:language", getGaussian(DC_LANGUAGE));
        ret.put("dc:coverage", getGaussian(DC_COVERAGE));
        ret.put("dc:source", getGaussian(DC_SOURCE));
        return ret;
    }

    protected String getGaussian(String[] words) {
        double g = Math.abs(rand.nextGaussian() / 4);
        g = Math.min(g, 1);
        int i = (int) Math.floor(g * (words.length - 1));
        return words[i];
    }

}
